# Copyright (C) 2021 Chi-kwan Chan
# Copyright (C) 2021 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from os.path  import splitext

from matplotlib import pyplot as plt
from matplotlib import dates  as mdates

def plot_site(dfs, title=None, fname=None, **kwargs):

    fig, axes = plt.subplots(6, 1, figsize=(12,12), sharex=True)

    for i, df in enumerate(dfs):
        alpha = 1.0 if i == 0 else (1 - i/len(dfs)) / 3
        width = 1.0 if i == 0 else 0.5
        for j, ax in enumerate(axes):
            ax.plot(df.date, df.iloc[:,j+1].fillna(0),
                    alpha=alpha, linewidth=width, **kwargs)
            ax.axvline(x=datetime.utcnow(), linestyle=':', color='k')

    for j, ax in enumerate(axes):
        ax.grid(alpha=0.25)
        ax.tick_params(axis='x',direction="in",top=True)
        ax.tick_params(axis='y',direction="in",right=True)
        ax.xaxis.set_major_locator(mdates.DayLocator())
        ax.autoscale(enable=True, axis='x', tight=True)

    axes[0].set_ylabel(r'$\tau_{255}$',      fontsize=16)
    axes[1].set_ylabel(r'$T_b$ [K]',         fontsize=16)
    axes[2].set_ylabel(r'pwv [mm]',          fontsize=16)
    axes[3].set_ylabel(r'lwp [kg m$^{-2}$]', fontsize=16)
    axes[4].set_ylabel(r'iwp [kg m$^{-2}$]', fontsize=16)
    axes[5].set_ylabel(r'o3 [DU]',           fontsize=16)

    axes[0].set_ylim(0, 1.1)
    axes[1].set_ylim(0, 330)
    axes[2].set_ylim(0, 16.5)
    axes[3].set_ylim(0, 2.2)
    axes[4].set_ylim(0, 2.2)
    axes[5].set_ylim(240, 460)

    if title is not None:
        axes[0].set_title(title, fontsize=16)

    axes[5].set_xlabel('Date', fontsize=16)
    axes[5].set_xlim(dfs[-1].date[0], None)
    plt.xticks(rotation=45, ha='right')

    fig.tight_layout()
    fig.subplots_adjust(wspace=0, hspace=0.05)

    if fname is not None:
        _, ext = splitext(fname)
        if ext[1:] in plt.gcf().canvas.get_supported_filetypes():
            fig.savefig(fname)
        else:
            fig.savefig(fname+'.pdf')
            fig.savefig(fname+'.png')


def plot_all(dfs, sites, title=None, fname=None, **kwargs):

    fig, axes = plt.subplots(6, 1, figsize=(12,12), sharex=True)

    for i, df in enumerate(dfs):
        if i == 0:
            kwa = {**kwargs, 'color':'dimgray'}
        else:
            kwa = kwargs
        for j, ax in enumerate(axes):
            ax.plot(df.date, df.iloc[:,j+1].fillna(0),
                    label=sites[i], **kwargs)
            ax.axvline(x=datetime.utcnow(), linestyle=':', color='k')

    for j, ax in enumerate(axes):
        ax.grid(alpha=0.25)
        ax.tick_params(axis='x',direction="in",top=True)
        ax.tick_params(axis='y',direction="in",right=True)
        ax.xaxis.set_major_locator(mdates.DayLocator())
        ax.autoscale(enable=True, axis='x', tight=True)

    axes[0].set_ylabel(r'$\tau_{255}$',      fontsize=16)
    axes[1].set_ylabel(r'$T_b$ [K]',         fontsize=16)
    axes[2].set_ylabel(r'pwv [mm]',          fontsize=16)
    axes[3].set_ylabel(r'lwp [kg m$^{-2}$]', fontsize=16)
    axes[4].set_ylabel(r'iwp [kg m$^{-2}$]', fontsize=16)
    axes[5].set_ylabel(r'o3 [DU]',           fontsize=16)

    axes[0].set_ylim(0, 1.1)
    axes[1].set_ylim(0, 330)
    axes[2].set_ylim(0, 16.5)
    axes[3].set_ylim(0, 2.2)
    axes[4].set_ylim(0, 2.2)
    axes[5].set_ylim(240, 460)

    if title is not None:
        axes[0].set_title(title, fontsize=16)

    axes[4].legend(loc='upper right', bbox_to_anchor=(0.99,2))

    axes[5].set_xlabel('Date', fontsize=16)
    axes[5].set_xlim(dfs[-1].date[0], None)
    plt.xticks(rotation=45, ha='right')

    fig.tight_layout()
    fig.subplots_adjust(wspace=0, hspace=0.05)

    if fname is not None:
        _, ext = splitext(fname)
        if ext[1:] in plt.gcf().canvas.get_supported_filetypes():
            fig.savefig(fname)
        else:
            fig.savefig(fname+'.pdf')
            fig.savefig(fname+'.png')



# Copyright (C) 2021 Chi-kwan Chan
# Copyright (C) 2021 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from random   import randrange
from os       import symlink, rename, path
from datetime import timedelta
from math     import sqrt

import requests
import pandas as pd
from tqdm import tqdm

import ucast  as uc
from ucast.io import dt_fmt

columns   = ['date', 'tau', 'Tb', 'pwv', 'lwp', 'iwp', 'o3']
forecasts = list(range(120+1)) + list(range(123, 384+1, 3))


am = uc.am.AM()


def ucast_dataframe(site, cycle):
    df = pd.DataFrame(columns=columns)

    for hr_forecast in tqdm(forecasts, desc=cycle.strftime(dt_fmt)):
        try:
            gfs = uc.gfs.GFS(site, cycle, hr_forecast)
        except requests.exceptions.RetryError as e:
            continue # skip a row

        sol  = am.solve(gfs)
        date = (gfs.cycle + timedelta(hours=hr_forecast)).strftime(dt_fmt)
        df   = df.append({'date':date, **sol}, ignore_index=True)

    return df


def valid(fname):
    if path.isfile(fname):
        with open(fname) as f:
            return len(f.readlines()) == len(forecasts)+1
    return False


def forced_symlink(src, dst):
    r   = randrange(1000000)
    tmp = f"{dst}-tmp{r:06d}"
    symlink(path.relpath(src, path.dirname(tmp)), tmp)
    rename(tmp, dst)


def regroup(sites):

    def d(s1, s2):
        dlat = s1.lat - s2.lat
        dlon = s1.lon - s2.lon
        return sqrt(dlat * dlat + dlon * dlon)

    ss = [getattr(uc.site, s) for s in sites]

    m = {}
    for i, curr in enumerate(ss):
        added = False
        for prev in ss[:i]:
            if d(curr, prev) < 0.1:
                added = True
                m[prev.name] = ','.join((m[prev.name], curr.name))
                break
        if not added:
            m[curr.name] = curr.name

    return m



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from tempfile import NamedTemporaryFile

from ...request import request
from .nomads    import data_url
from .grib      import load

GRIDSZ = 0.25  # Grid spacing string: the available GFS lat,lon grid
               # spacings are 0.25, 0.50, or 1.00 degrees.

class GFS:

    def __init__(self, site, cycle, product=None, gridsz=GRIDSZ):

        # Step 1: download data from NOMADS
        r = request(data_url(site, cycle, product, gridsz))

        # Step 2: save data to temporary file; load it back with `pygrib`
        with NamedTemporaryFile() as t:
            with open(t.name, "wb") as f:
                f.write(r.content)
            d = load(t.name, site)

        # Step 3: set the instance attributes
        self.site    = site
        self.cycle   = cycle
        self.product = product
        self.gridsz  = gridsz
        for k, v in d.items():
            setattr(self, k, v)



# Copyright (C) 2021 Chi-kwan Chan
# Copyright (C) 2021 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

import numpy  as np
import pandas as pd

names   = [             'date',    'tau',       'Tb',      'pwv',  'lwp',       'iwp',             'o3']
heading = "#            date       tau225        Tb[K]      pwv[mm] lwp[kg*m^-2] iwp[kg*m^-2]       o3[DU]\n"
out_fmt = "%16s %12.4e %12.4e %12.4e %12.4e %12.4e %12.4e"
dt_fmt  = "%Y-%m-%d_%H.%M.%S"


def save_txt(fname, df):
    with open(fname, "w") as f:
        f.write(heading)
        np.savetxt(f, df.fillna(0).values, fmt=out_fmt)

def read_txt(fname):
    d = pd.read_csv(fname, delim_whitespace=True, skiprows=1, names=names)
    d.date = pd.to_datetime(d.date, format=dt_fmt)
    return d


def save_csv(fname, df):
    df.to_csv(fname, index=False)

def read_csv(fname):
    d = pd.read_csv(fname)
    d.date = pd.to_datetime(d.date, format=dt_fmt)
    return d


def save_tsv(fname, df):
    df.to_csv(fname, index=False, sep='\t')

def read_tsv(fname):
    d = pd.read_csv(fname, sep='\t')
    d.date = pd.to_datetime(d.date, format=dt_fmt)
    return d



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ...weather.gfs.nomads import levels

# Physical constants
G_STD               = 9.80665  # standard gravity [m / s^2]
M_AIR               = 28.964   # average dry air mass [g / mole]
M_O3                = 47.997   # O3 mass [g / mole]
H2O_SUPERCOOL_LIMIT = 238.     # Assume ice below this temperature [K]
PASCAL_ON_MBAR      = 100.     # conversion from mbar (hPa) to Pa

RH_TOP_PLEVEL = 29.
STRAT_H2O_VMR = 5e-6

def column(name, value):
    fu_map = {
        'o3 vmr'          :('.3e', ''        ),
        'h2o vmr'         :('.3e', ''        ),
        'h2o RH'          :('.2f', '%'       ),
        'h2o RHi'         :('.2f', '%'       ),
        'lwp_abs_Rayleigh':('.3e', ' kg*m^-2'),
        'iwp_abs_Rayleigh':('.3e', ' kg*m^-2'),
    }
    if value > 0:
        fmt, unit = fu_map[name]
        return f"column {name} {value:{fmt}}{unit}"
    else:
        return None

def layer(Pb, zb, Tb, T, o3_vmr, RH, ctw, cti):
    if T > H2O_SUPERCOOL_LIMIT:
        h2o_label = "h2o RH"
        ctw_label = "lwp_abs_Rayleigh"
    else:
        # Below the supercooling limit, assume any liquid water
        # is really ice.  (GFS 15 occasionally has numerically
        # negligible amounts of liquid water at unphysically
        # low temperature.)
        h2o_label = "h2o RHi"
        ctw_label = "iwp_abs_Rayleigh"

    if (Pb > RH_TOP_PLEVEL):
        h2o_column = column(h2o_label, RH)
    else:
        # Above the level defined above, use a fixed stratospheric
        # mixing ratio instead of RH from GFS.
        h2o_column = column("h2o vmr", STRAT_H2O_VMR)

    return '\n'.join(filter(None, [
        "layer",
       f"Pbase {Pb:.1f} mbar  # {zb:.1f} m",
       f"Tbase {Tb:.1f} K",
        "column dry_air vmr",
        column("o3 vmr", o3_vmr),
        h2o_column,
        column(ctw_label, ctw),
        column("iwp_abs_Rayleigh", cti),
    ]))

def interpos(a, b, u, log=False):
    if log:
        return np.exp(np.log(a) + np.log(b/a) * u)
    else:
        c = a + (b - a) * u
        return c if c > 0.0 else 0.0

def base(arr, n, u, log=False):
    if u == 0.0:
        return arr[:n]
    else:
        return np.r_[arr[:n], interpos(arr[n-1], arr[n], u, log=log)]

def average(arr, n, u):
    arr = base(arr, n, u)
    return np.r_[arr[0], 0.5 * (arr[:-1]+arr[1:])]

def delta(arr, n, u):
    arr = base(arr, n, u)
    return np.r_[arr[0], arr[1:]-arr[:-1]]

def config(gfs, debug=False):

    p = f"{gfs.product:03d} hour forecast" if isinstance(gfs.product, int) else "analysis"
    z = gfs.site.alt

    # Prepare for regriding
    if z > gfs.z[0]:
        raise ValueError("User-specified altitude exceeds top GFS level")

    # For pressure, height, and temperature, use the base values.
    # The surface (last) values are interpolated with respect to z
    n = np.argmax(gfs.z < z)
    u = (gfs.z[n-1]-z) / (gfs.z[n-1]-gfs.z[n])

    Pb = base(gfs.P, n, u, log=True)
    zb = base(gfs.z, n, u)
    Tb = base(gfs.T, n, u)

    # For mixing ratios and RH, use averages over the two levels
    # bounding the layer.
    # The surface (last) value is interpolated with respect to P
    u = (gfs.P[n-1]-Pb[-1]) / (gfs.P[n-1]-gfs.P[n])

    T         = average(gfs.T,         n, u)
    o3_mmr    = average(gfs.o3_mmr,    n, u)
    RH        = average(gfs.RH,        n, u)
    cloud_lmr = average(gfs.cloud_lmr, n, u)
    cloud_imr = average(gfs.cloud_imr, n, u)

    # Convert cloud liquid water mixing ratios (lmr) and ice mixing
    # ratio (imr) [kg / kg] to cloud total liquid water (ctw) and
    # cloud total ice (cti) across the layer [kg / m^2].
    m   = delta(Pb, n, u) * (PASCAL_ON_MBAR/G_STD)
    ctw = cloud_lmr * m
    cti = cloud_imr * m

    # Convert mass mixing ratio to volume mixing ratio
    o3_vmr = o3_mmr * (M_AIR/M_O3)

    # Dump arrays to debug
    if debug:
        import pandas as pd
        loc = locals()
        org = pd.DataFrame({k:getattr(gfs, k) for k in [
            'P', 'z', 'T', 'o3_mmr', 'RH', 'cloud_lmr', 'cloud_imr']})
        der = pd.DataFrame({k:loc[k] for k in [
            'Pb', 'zb', 'Tb', 'T', 'o3_vmr', 'RH', 'cloud_lmr', 'cloud_imr', 'm', 'ctw', 'cti']})
        org.to_csv('debug-org.csv', index=False)
        der.to_csv('debug-der.csv', index=False)

    # Finally construct the configuration text
    l = [f"""#
# Layer data below were derived from NCEP GFS model data obtained
# from the NOAA Operational Model Archive Distribution System
# (NOMADS).  See http://nomads.ncep.noaa.gov for more information.
#
#         Production date: {gfs.cycle:%Y%m%d}
#                   Cycle: {gfs.cycle:%H} UT
#                 Product: {p}
#
# Interpolated to
#
#                latitude: {gfs.site.lat} deg. N
#               longitude: {gfs.site.lon} deg. E
#   Geopotential altitude: {z} m
#"""]

    for i in range(len(T)):
        l.append(layer(
            Pb[i], zb[i], Tb[i],
            T[i], o3_vmr[i], RH[i], ctw[i], cti[i]))

    return "\n\n".join(l)



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from os       import path
from glob     import glob
from datetime import datetime

import requests
import click

import ucast as uc
from ucast.utils import forced_symlink  as symlink
from ucast.utils import ucast_dataframe as mkdf
from ucast.utils import forecasts, valid, regroup
from ucast.io    import dt_fmt
from ucast.io    import save_tsv as save
from ucast.io    import read_tsv as read
from ucast.plot  import plot_site, plot_all
from ucast.bokeh import static_vis


@click.group()
def ucast():
    """µcast: micro-weather forecasting for astronomy"""


@ucast.command()
@click.argument("site")
@click.option("--lag",     default=5.25,  help="Lag hour for weather forecast.")
@click.option("--data",    default=None,  help="Data archive directory.")
@click.option("--link",    default=None,  help="Directory with latest links.")
@click.option("--no-link", default=False, help="Disable latest links.", is_flag=True)
def mktab(lag, site, data, link, no_link):
    """Pull weather for telescope SITE, process with `am`, and make tables """

    if no_link and link is not None:
        raise click.UsageError(
            '"--link" should not be specified if "--no-link" is set')

    if data is None:
        data = site if path.isdir(site) else '.'

    if link is None:
        link = data

    site         = getattr(uc.site, site)
    latest_cycle = uc.gfs.latest_cycle(lag=lag)

    for hr_ago in range(0, 48+1, 6):
        cycle   = uc.gfs.relative_cycle(latest_cycle, hr_ago)
        outfile = path.join(data, cycle.strftime(dt_fmt)+'.tsv')

        if valid(outfile):
            print(f'Skip "{outfile}"', end='')
        else:
            print(f'Creating "{outfile}" ...', end='')
            save(outfile, mkdf(site, cycle))
            print(" DONE", end='')

        if no_link:
            print()
        else:
            target = path.join(link,
                "latest.tsv" if hr_ago == 0 else f"latest-{hr_ago:02d}.tsv")
            print(f'; linked as "{target}"')
            symlink(outfile, target)


@ucast.command()
@click.argument("site")
@click.option("--no-lag",  default=False, help="Do not use current time to name links.", is_flag=True)
@click.option("--lag",     default=None,  help="Lag hour for weather forecast.")
@click.option("--data",    default=None,  help="Data archive directory.")
@click.option("--link",    default=None,  help="Directory with latest links.")
def link(lag, no_lag, site, data, link):
    """(Re)link weather forecast tables"""

    if no_lag and lag is not None:
        raise click.UsageError(
            '"--lag" should not be specified if "--no-lag" is set')
    if lag is None:
        lag = 5.25

    if data is None:
        data = site if path.isdir(site) else '.'

    if link is None:
        link = data

    if no_lag:
        pattern = path.join(link, '????-??-??_??.??.??.tsv')
        sites = [p.split(path.sep)[-2] for p in sorted(glob(pattern))]

    else:
        latest_cycle = uc.gfs.latest_cycle(lag=lag)

    for hr_ago in range(0, 48+1, 6):
        cycle   = uc.gfs.relative_cycle(latest_cycle, hr_ago)
        outfile = path.join(data, cycle.strftime(dt_fmt)+'.tsv')

        if valid(outfile):
            target = path.join(link,
                "latest.tsv" if hr_ago == 0 else f"latest-{hr_ago:02d}.tsv")
            print(f'"{outfile}" linked as "{target}"')
            symlink(outfile, target)
        else:
            print(f'"{outfile}" is missing; skipped')


@ucast.command()
@click.argument("site")
@click.option("--link", default=None, help="Directory with latest links.")
@click.option("--out",  default=None, help="File name of the plot.")
def psite(site, link, out):
    """Read weather tables from SITE and create summary plot for one site"""

    if link is None:
        link = site if path.isdir(site) else '.'

    if out is None:
        out = path.join(link, 'latest')
    elif '/' not in out:
        out = path.join(link, out)

    site         = getattr(uc.site, site)
    latest_fname = path.basename(path.realpath(path.join(link, "latest.tsv")))
    latest_cycle = datetime.strptime(path.splitext(latest_fname)[0], dt_fmt)

    dfs = []
    for hr_ago in range(0, 48+1, 6):
        target = path.join(link,
            "latest.tsv" if hr_ago == 0 else f"latest-{hr_ago:02d}.tsv")
        dfs.append(read(target))

    title = f'{site.name}: ({site.lat},{site.lon},{site.alt})@{latest_cycle}'
    plot_site(dfs, title, fname=out, color='k')


@ucast.command()
@click.argument("sites", nargs=-1)
@click.option("--link", default=None, help="Directory with latest links.")
@click.option("--set",  default=None, help="Input dataset, e.g. latest, latest-06, ...")
@click.option("--out",  default=None, help="File name of the plot.")
def pall(sites, link, set, out):
    """Read weather tables from SITES and create summary plot for all sites"""

    if link is None:
        link = '.'

    if set is None:
        set = 'latest'

    if out is None:
        out = path.join(link, set)
    elif '/' not in out:
        out = path.join(link, out)

    if len(sites) == 0:
        pattern = path.join(link, "*", set+'.tsv')
        sites = [p.split(path.sep)[-2] for p in sorted(glob(pattern))]
        if len(sites) == 0:
            print('Weather table not found.')
            return 0

    dfs = [read(f'{s}/{set}.tsv') for s in sites]
    plot_all(dfs, sites, fname=out)


@ucast.command()
@click.argument("sites", nargs=-1)
@click.option("--link", default=None, help="Directory with latest links.")
@click.option("--set",  default=None, help="Input dataset, e.g. latest, latest-06, ...")
@click.option("--out",  default=None, help="File name of the plot.")
@click.option('--browser/--no-browser',
                        default=True, help="Open visualization in a browser.")
def vis(sites, link, set, out, browser):
    """ Creates a bokeh html file containing forecast of all sites."""

    if link is None:
        link = '.'

    if set is None:
        set = 'latest'

    if out is None:
        out = path.join(link, set)
    elif '/' not in out:
        out = path.join(link, out)

    if len(sites) == 0:
        pattern = path.join(link, "*", set+'.tsv')
        sites = [p.split(path.sep)[-2] for p in sorted(glob(pattern))]
        if len(sites) == 0:
            print('Weather table not found.')
            return 0

    sites = regroup(sites)
    dfs   = [read(f'{s}/{set}.tsv') for s in sites]
    sites = list(sites.values())
    static_vis(dfs, sites, fname=out, browser=browser)


if __name__ == "__main__":
    ucast()



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

import shutil
import subprocess
import re

from .config import config

class AM:

    # Column density units (cm^-2 equivalents)
    MM_PWV   = 3.3427e21
    KG_ON_M2 = 3.3427e21
    DU       = 2.6868e16

    header = """#
# This header is prepended to the atmospheric layers generated by
# gfs16_to_am10.py to generate a complete am configuration file.
#
f 225 GHz 225 GHz 1 GHz
output f GHz tau Tb K
T0 2.7 K
"""

    cmap = {
        'pwv': (re.compile('^#.*h2o'             ), MM_PWV  ),
        'lwp': (re.compile('^#.*lwp_abs_Rayleigh'), KG_ON_M2),
        'iwp': (re.compile('^#.*iwp_abs_Rayleigh'), KG_ON_M2),
        'o3' : (re.compile('^#.*o3'              ), DU      ),
    }

    def __init__(self, path=None):
        if path is None:
            path = shutil.which('am')
        if path is None:
            raise OSError('Executable of `am` Atmospheric Model not found.')

        self.am = path

    def run(self, *args, **kwargs):
        return subprocess.run([self.am, *args], **kwargs)

    def solve(self, gfs):
        cfg = self.header + config(gfs)
        res = self.run('-', input=cfg, encoding='utf-8', capture_output=True)
        sol = {}

        l = res.stdout.split()
        sol['tau'] = float(l[1])
        sol['Tb' ] = float(l[2])

        for l in res.stderr.split('\n'):
            for k, (p, f) in self.cmap.items():
                if p.match(l):
                    sol[k] = float(l.split()[2]) / f
                    continue

        return sol



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple

Site = namedtuple('Site', ['name', 'lat', 'lon', 'alt'])

ALMA = Site('ALMA', -23.029,  -67.755, 5070.4)
APEX = Site('APEX', -23.006,  -67.759, 5104.5)
GLT  = Site('GLT',   72.579,  -38.454, 3204.9)
JCMT = Site('JCMT',  19.823, -155.477, 4120.1)
KP   = Site('KP',    31.956, -111.612, 1902.0)
LMT  = Site('LMT',   18.985,  -97.315, 4600.0)
PDB  = Site('PDB',   44.634,    5.907, 2616.4)
PV   = Site('PV',    37.066,   -3.393, 2921.7)
SMA  = Site('SMA',   19.822, -155.476, 4110.3)
SMT  = Site('SMT',   32.702, -109.891, 3158.7)
SPT  = Site('SPT',  -90.000,   45.000, 2857.4)



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from math import floor

import sys
import requests
import time

# GFS variables to be requested
variables = (
    "HGT",   # Geopotential height [m]
    "TMP",   # Temperature [K]
    "O3MR",  # Ozone mass mixing ratio [kg O3 / kg air]
    "RH",    # Relative Humidity [%]
    "CLWMR", # Cloud liquid water mass mixing ratio [kg liquid / kg air]
    "ICMR",  # Cloud ice mass mixing ratio [kg liquid / kg air]
)

# GFS grid levels [mbar]
levels = (
    1, 2, 3, 5, 7, 10, 20, 30, 50, 70, 100, 150, 200, 250, 300, 350, 400,
    450, 500, 550, 600, 650, 700, 750, 800, 850, 900, 925, 950, 975, 1000,
)

def cgi_url(g):
    """URL for the CGI interface for getting GFS data."""
    return f"https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_{g}_1hr.pl"

def product_query(cycle, g, p):
    """Query string for requesting the kind of data product."""
    return f"file=gfs.t{cycle:%H}z.pgrb2.{g}.{p}"

def variable_query(v):
    """Query string for adding GFS variables to the CGI request URL."""
    return f"var_{v}=on"

def level_query(l):
    """Query string used to add GFS grid level to the CGI request URL."""
    return f"lev_{l:d}_mb=on"

def subregion_query(lat, lon, gridsz):
    """Query string for the grid subset request."""
    l = floor(lon / gridsz) * gridsz
    b = floor(lat / gridsz) * gridsz
    r = l + gridsz
    t = b + gridsz
    return f"subregion=&leftlon={l}&rightlon={r}&toplat={t}&bottomlat={b}"

def cycle_query(cycle):
    """Query string for requesting the specific data and production cycle"""
    return f"dir=%2Fgfs.{cycle:%Y%m%d}%2F{cycle:%H}%2Fatmos"

def data_url(site, cycle, product, gridsz):
    """Construct the full data request URL.

    These include the base URL for the NOMADS CGI, and various strings
    for formatting the arguments given to it.  Note that some
    information (e.g. grid, forecast production cycle) gets used more
    than once to construct the CGI request.

    """
    # Forecast product: either "anl" for analysis at production
    # time, or "fxxx" for forecast xxx hours in the future, where
    # xxx ranges from 000 to 384 by 1-hour steps, by 1-hour steps
    # up to 120 hours, and by 3-hour steps thereafter.
    p = f"f{product:03d}" if isinstance(product, int) else "anl"

    # In the GFS file names and CGI interface, this is coded as "0p25"
    # for 0.25 deg, etc.
    g = f"{gridsz:.2f}".replace('.', 'p')

    query = '&'.join([
        product_query(cycle, g, p),
        '&'.join(level_query(l)    for l in levels),
        '&'.join(variable_query(v) for v in variables),
        subregion_query(site.lat, site.lon, gridsz),
        cycle_query(cycle),
    ])
    return '?'.join([cgi_url(g), query])



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pygrib

from .nomads import levels

BADVAL = -99999.0  # placeholder for missing or undefined data

load_map = {
    'z'        : ("Geopotential Height", BADVAL),
    'T'        : ("Temperature",         BADVAL),
    'o3_mmr'   : ("Ozone mixing ratio",     0.0),
    'RH'       : ("Relative humidity",      0.0),
    'cloud_lmr': ("Cloud mixing ratio",     0.0),
    'cloud_imr': ("Ice water mixing ratio", 0.0),
}

def load(path, site, grid_delta=0.25):

    u = (site.lat/grid_delta) % 1
    v = (site.lon/grid_delta) % 1

    idx = pygrib.index(path, "name", "level")

    def grid_interp(name, level, bad):
        try:
            a = idx.select(name=name, level=level)[0].values
        except:
            return bad
        else:
            return (a[0][0] * (1.0-u) * (1.0-v) + a[1][0] * u * (1.0-v) +
                    a[0][1] * (1.0-u) *      v  + a[1][1] * u *      v   )

    d = {'P':np.array(levels)}
    for k, (name, bad) in load_map.items():
        d[k] = np.array([grid_interp(name, l, bad) for l in levels])

    return d



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta

def relative_cycle(ref, lag):
    """GFS forecast cycle time relative to another time.

    This function returns the datetime corresponding to a GFS forecast
    cycle time displaced by some number of hours relative to another
    GFS cycle time.  For example,

        relative_cycle(ref, 12)

    will return the GFS cycle time 12 hours before the cycle
    corresponding to `ref`

    Args:
        ref: Reference time
        lag: Lag time in hour

    Returns:
        The relative GFS cycle time.

    """
    last = ref - timedelta(hours=lag)
    return last.replace(hour=(last.hour//6)*6,
                        minute=0,
                        second=0,
                        microsecond=0)

def latest_cycle(lag=6):
    """The most recent GFS cycle time.

    Args:
        lag: The default GFS forecast production lag is 6 hours.  To
            get an earlier forecast, this might be set somewhat
            shorter, with due care taken to ensure it isn't set too
            short.

    Returns:
        The most recent GFS cycle time.

    """
    return relative_cycle(datetime.utcnow(), lag)



#!/usr/bin/env python3
#
# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

setup(
    name='ucast',
    version='0.1.0',
    url='https://github.com/focisrc/ucast',
    author='Chi-kwan Chan',
    author_email='chanc@arizona.edu',
    description='Micro-weather forecasting for astronomy',
    packages=find_packages('mod'),
    package_dir={'': 'mod'},
    entry_points={
        'console_scripts': [
            'ucast = ucast.__main__:ucast',
        ],
    },
    python_requires='>=3.6', # `ucast` uses python3's f-string and typing
    install_requires=[
        'click>=7.1.2',
        'matplotlib>=3.2.2',
        'numpy>=1.18.5',
        'pandas>=1.0.5',
        'pygrib>=2.0.4',
        'requests>=2.24.0',
        'tqdm>=4.46.1',
        'bokeh>=2.3.0'
    ],
)



from bokeh.models import Range1d, Column
from bokeh.plotting import figure, output_file, show, save

from bokeh.palettes import Category10_9 as palette
import itertools

colors = itertools.cycle(palette)
tools  = "pan,box_zoom,wheel_zoom,box_select,undo,redo,reset,save".split()

def static_vis(dfs, sites, fname, browser):
    output_file(fname+'.html')

    plots = []
    for var in "tau,pwv,lwp,iwp".split(","):
        if len(plots) == 0:
            kwargs = {}
        else:
            kwargs = {'x_range':plots[0].x_range}
        plots.append(create_plot(var, dfs, sites, **kwargs))

    c = Column(*plots)
    if browser:
        show(c)
    else:
        print(f'Save dashboard to "{fname}.html"')
        save(c)


def create_plot(var, dfs, sites, **kwargs):
    p = figure(title=var, x_axis_type="datetime",
               plot_width=1600,
               plot_height=500 if var == "tau" else 300,
               tools=tools,
               **kwargs)

    if var == "tau":
        p.y_range = Range1d(0, 1)
    elif var == "pwv":
        p.y_range = Range1d(0, 15)
    elif var == "lwp":
        p.y_range = Range1d(0, 2)
    elif var == "iwp":
        p.y_range = Range1d(0, 2)
    for i, (df, c) in enumerate(zip(dfs, colors)):
        p.line(df['date'], df[var].fillna(0),
               line_width=2,
               color=c, legend_label=sites[i],
               alpha=0.75, muted_alpha=0.25)

    p.legend.location = "top_right"
    p.legend.click_policy = "mute"

    return p



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of `ucast`.
#
# `Ucast` is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# `Ucast` is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with `ucast`.  If not, see <http://www.gnu.org/licenses/>.

import sys
import time
import requests

# Timeouts and retries
CONN_TIMEOUT        = 6   # Initial server response timeout in seconds
READ_TIMEOUT        = 18  # Stalled download timeout in seconds
RETRY_DELAY         = 60  # Delay before retry (NOAA requests 60 s)
MAX_DOWNLOAD_TRIES  = 5

def errln(s):
    print(s, file=sys.stderr)

def err(s):
    print(s, file=sys.stderr, end='')

def request(url,
            ctime=CONN_TIMEOUT,
            rtime=READ_TIMEOUT,
            delay=RETRY_DELAY,
            retry=MAX_DOWNLOAD_TRIES):
    while True:
        retry -= 1

        try:
            r = requests.get(url, timeout=(ctime, rtime))
            if r.status_code == requests.codes.ok:
                return r
        except requests.exceptions.ConnectTimeout:
            err("Connection timed out.")
        except requests.exceptions.ReadTimeout:
            err("Data download timed out.")
        else:
            err(f"Download failed with status code {r.status_code}.")

        if retry:
            errln("  Retrying...")
            time.sleep(delay)
        else:
            errln("  Giving up.")
            break

    errln(f'Failed URL was: "{url}".')
    raise requests.exceptions.RetryError(
        f'Tried downloading "{url}" {retry} times but failed')
